#
#  @file   CpaStates.py
#  @author Francis Moreau
#  date    Mon May 22 2007
#  
#  Description:
#
#  Session Controller's states when it is communicating with the CPA.
#
#  Copyright (C) 2007 Paraxip Technologies.
#  All Rights Reserved.
#

#
# Our imports
#

#Logging
import paraxip.infra.PyLoggingFunctions as PyLoggingFunctions
from netborderPyInfra import TraceScope, CallLogger, Logger, PyString

#@tag FIX THE import *, use from <x> import <y>
from PyResipStack import *
from SipCallEngineExtending import *

from paraxip.sipua.internal.SipCallEngineState import *
from paraxip.sipua.internal.SipCallEngineEvent import *
from paraxip.sipua.internal.ConnectionStatus import *
from paraxip.sipua.internal.SMTimeoutProcessor import *

from paraxip.sipua.user.b2bua import B2buaStates

from paraxip.sipua.user.cpa import SMTimerManager
from paraxip.sipua.user.cpa import CpaResult
from paraxip.sipua.user.cpa import DetailedCpdResult
from paraxip.sipua.user.cpa import CallEvent

# CPASTATS
from paraxip.sipua.ui.model.cpaOM import CpaOM 

#
# Python core modules imports
#
import time
import re
import datetime

# Force declaration of class attributes in __init__ method
__pychecker__ = "initattr"

cpaStatesProcessEventLogger = Logger('netborder.processevent.cpa')

class CpaStates(SipCallEngineState):
  """
  Logged as: netborder.cpa.CpaCall
  """
  
  def __init__(self, in_stateMachine):
    SipCallEngineState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CpaStates.__init__")
  
  def processEvent(self, in_event, out_pystrNewStateName):
    
    PyLoggingFunctions.log_info(cpaStatesProcessEventLogger,
                "CpaStates processEvent(in_event= %s, out_pystrNewStateName=%s)." %
                (in_event, out_pystrNewStateName))
    
    #Could do the following to automatically call the 'right, virtual' function without
    #having to do the endless ifinstance(...). Plus this way would be truly polymorphic,
    #not just overwriting the function.
    #handleFunctionName = '_handle%s' % in_event.__class__.__name__
    #rv = True
    #if hasattr(self, handleFunctionName):
    #  handleFn = getattr(self, handleFunctionName)
    #  rv = handleFn(in_event, out_pystrNewStateName)
    
    return self.processEvent_i(in_event, out_pystrNewStateName)
  
  def _handleCallerDisconnecting(self, in_event, out_pystrNewStateName):
    sipFailedStatusCode = 487
    self._stateMachine.notifyCallEvent( CallEvent.SipError( CallEvent.SipErrorOriginator.Caller, 
                                                            CallEvent.SipErrorCode(sipFailedStatusCode))
                                       )   
    out_pystrNewStateName.set("DISCONNECTING")
    
  def _handleCalleeDisconnecting(self, in_event, out_pystrNewStateName):
    sipFailedStatusCode = 603
    self._stateMachine.notifyCallEvent( CallEvent.SipError( CallEvent.SipErrorOriginator.Callee, 
                                                            CallEvent.SipErrorCode(sipFailedStatusCode))
                                       )
    out_pystrNewStateName.set("DISCONNECTING")
    
  def _handleCpaDisconnecting(self, in_event, out_pystrNewStateName):
    # CPA UAS hung-up unexpectedly
    self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnected
    
    if (self._stateMachine.calleeLegStatus == ConnectionStatus.Connected):
      # Re-invite the callee to minimize damage
      #@tag ???
      #PyLoggingFunctions.log_info(self.getLogger(),
      #  "CpaStates CPA "
      #  "endpoint disconnected unexpectedly")
      out_pystrNewStateName.set("REINVITING_CALLEE")
      return True
    else:
      # Stay in the same state and let the connection with the callee being
      # established. If the callee can't establish the connection within
      # the time remaining to the timer then the timer will expired and
      # everything will be managed in the TimeoutEvent of this state.
  
      # TODO optionally reject caller and cancel callee
      return True
  
  def _handleDialogToCallTerminatedEvent(self, in_event, out_pystrNewStateName):
    if (in_event.getDialogId() == self._stateMachine.callerDialogId or
        in_event.getDialogId() == self._stateMachine.calleeDialogId):
      self._stateMachine.dialogIdInitiatingDisconnect = \
        in_event.getDialogId()
      
      #Flag the call with the proper error code.
      if in_event.getDialogId() == self._stateMachine.callerDialogId:
        self._handleCallerDisconnecting(in_event, out_pystrNewStateName)
      elif in_event.getDialogId() == self._stateMachine.calleeDialogId:
        self._handleCalleeDisconnecting(in_event, out_pystrNewStateName)
      else:
        PyLoggingFunctions.log_error(\
            self.getLogger(), 
            "Invalid DialogToCallTerminatedEvent, dialog ID is neither caller or callee.")
     
      return True
      
    elif (in_event.getDialogId() == self._stateMachine.cpaDialogId):
      self._handleCpaDisconnecting(in_event, out_pystrNewStateName)
      return True

########################################################################

class DialingCpaState(CpaStates):
  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DialingCpaState.__init__")
    self._recvOnlyRegExp = re.compile('recvonly')
    self._mediaLineRegExp = re.compile("m=.*")
  
  def getName(self):
    return "DIALING_CPA"
  
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)

    # Inviting cpa endpoint without sdp
    self._stateMachine.callCpaEndpoint()

    out_pystrNewStateName.set(self.getName())

  def processEvent_i(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, OutDialogToCallCallingEvent)):
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallFailedEvent)): #OUT
      self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnecting
      
      eventSipMessage = in_event.getSipMessage()
      strWhyFailed = eventSipMessage.getReasonText()
      
      sipFailStatusCode = in_event.getSipMessage().getStatusCode()
      
      # There is a problem connecting with the CPA UAS
      self._stateMachine.uiCpaSipErrorCode = sipFailStatusCode        
      
      #
      # Sometimes: 'Decline (Licensing capacity has been reached).'
      #
      self._stateMachine.notifyCallEvent( CallEvent.SipError( CallEvent.SipErrorOriginator.Cpa, 
                                                              CallEvent.SipErrorCode(sipFailStatusCode),
                                                              strWhyFailed) )
      
      # Reject the incoming call right away
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, OutDialogToCallProvisionalEvent)):
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent)): #OUT
      # Got an SDP offer in response to our INVITE w/o SDP
      # 1. Save the sdp offered by cpa endpoint
      self._stateMachine.cpaSdpAnswered = in_event.getSdpContents()
      
      # 2. Provide a ACK with a dummy SDP to the CPA Media UAS 
      # TODO: Should match m line/rtpmap of offer
      # TODO: SDP should be built only once
      errStr = PyString()
      tmpSdpAnswer = SDP()
      strSDP = "v=0\r\n" \
               "o=- 1182439127 1182439127 IN IP4 127.0.0.1\r\n" \
               "s=Netborder Call Processor Dummy Media Session\r\n" \
               "c=IN IP4 127.0.0.1\r\n" \
               "t=0 0\r\n" \
               "m=audio 9000 RTP/AVP 0\r\n" \
               "a=sendonly\r\n" \
               "a=rtpmap:0 PCMU/8000\r\n"

      if not tmpSdpAnswer.parse(strSDP, errStr):
        PyLoggingFunctions.log_error(self.getLogger(), \
            "failed to parse sdp :\n%s", errStr)
        return False

      if (not self._stateMachine.provideSdpAnswerToDialog(
            self._stateMachine.cpaDialogId, tmpSdpAnswer)):
        return False
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallConnectedEvent)):
      self._stateMachine.cpaLegStatus = ConnectionStatus.Connected
      
      # Inviting the callee with cpa endpoint sdp
      if not self._stateMachine.callCallee(self._stateMachine.cpaSdpAnswered):
        # callCallee is responsible for logging the error
        out_pystrNewStateName.set("DISCONNECTING")
        return True
      
      out_pystrNewStateName.set("DIALING_CALLEE")
      return True
    
    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      return self._handleDialogToCallTerminatedEvent(in_event, out_pystrNewStateName)
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

########################################################################

class DialingCalleeState(CpaStates):
  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DialingCalleeState.__init__")
    
    self._preConnectTimerId = None
  
  def getName(self):
    return "DIALING_CALLEE"
  
  def startPreConnectTimer(self):
    timeoutEvent = TimeoutEvent()
    timeoutCancelledEvent = TimeoutCancelledEvent()
    
    preConnectTimeoutProcessor = SMTimeoutProcessor(
      self._stateMachine,
      timeoutEvent,
      timeoutCancelledEvent)
    
    preConnectCpaResultTimeoutMs = \
      self._stateMachine.config[self._stateMachine.PARAM_CPA_PRECONNECT_TIMEOUT_MS]
    self._preConnectTimerId = self._stateMachine.globalContext.scheduleTimer(
      preConnectTimeoutProcessor,
      preConnectCpaResultTimeoutMs, 0)
    timeoutEvent.timerId = self._preConnectTimerId
    timeoutCancelledEvent.timerId = self._preConnectTimerId
  
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    self._preConnectTimerId = None
    self._stateMachine.reinvitedCpa = False
    self.startPreConnectTimer()
    
    # CPASTATS
    # Store CPA statistics: date and time dialing
    now=datetime.datetime.now()
    strNowDate=now.strftime("%Y-%m-%d")     
    strNowTime='%s.%03d'%(now.strftime("%H:%M:%S") ,now.microsecond/1000)
    self._stateMachine.cpaStats.addOM(CpaOM.OM_CPA_DATE,strNowDate)
    self._stateMachine.cpaStats.addOM(CpaOM.OM_CPA_TIME_DIALING,strNowTime)

    out_pystrNewStateName.set(self.getName())
  
  def processEvent_i(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0 maxlines=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, OutDialogToCallCallingEvent)):
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallFailedEvent)): #OUT
      self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      
      sipFailedStatusCode = in_event.getSipMessage().getStatusCode()
      
      self._stateMachine.notifyCallEvent( CallEvent.SipError( CallEvent.SipErrorOriginator.Callee, 
                                                              CallEvent.SipErrorCode(sipFailedStatusCode))
                                         )
        
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, OutDialogToCallEarlyMediaEvent)):
      if (in_event.getDialogId() == self._stateMachine.calleeDialogId):#SDP from callee
        # Provisional response with SDP
        if (self._stateMachine.getCallStack().hadSipError()):
          # CPA failed previously but we still want to connect caller and callee
          #   Wait for callee connecting to be able to do re-invite,
          #   do not change state
          return True
        else:
          # Bug Fix 1863: relay 183 w/SDP as user configurable provisional message
          #   (without SDP). Other provisional responses are relayed with same status
          #   code but without SDP
          statusCodeToRelay = None
          statusCode = in_event.getSipMessage().getStatusCode()
          if (statusCode == 183):
            statusCodeToRelay = self._stateMachine.config[
              self._stateMachine.PARAM_CPA_PROVISIONAL_RELAYED_UPON_183]
          else:
            statusCodeToRelay = statusCode

          self._stateMachine.sendProvisionalToInDialog(
            self._stateMachine.callerDialogId, statusCodeToRelay)

          if ((self._stateMachine.cpaDialogId is not None ) and
              (self._stateMachine.cpaLegStatus == ConnectionStatus.Connected)):
            # re-invite cpa in order to advertise calle's sdp
            calleeSdp =  in_event.getSdpContents()
            if self._stateMachine.reInviteDialog(self._stateMachine.cpaDialogId,
                                              calleeSdp):
              self._stateMachine.reinvitedCpa = True
            
          out_pystrNewStateName.set("WAITING_CPA_RESULT")


      return True
    
    elif (isinstance(in_event, OutDialogToCallProvisionalEvent)):
      if (in_event.getDialogId() == self._stateMachine.calleeDialogId):
        # Provisional response: relay if no SDP present, otherwise relay is handled
        #   by OutDialogToCallEarlyMediaEvent event handling above
        if not in_event.getSipMessage().hasContents():
          self._stateMachine.sendProvisionalToInDialog(self._stateMachine.callerDialogId, in_event.getSipMessage().getStatusCode())
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent)): #OUT

      if (in_event.getDialogId() == self._stateMachine.calleeDialogId):#SDP from callee

        if (self._stateMachine.getCallStack().hadSipError()):
          # Save the callee sdp answered
          self._stateMachine.calleeSdp = in_event.getSdpContents()
        else:

          if ((self._stateMachine.cpaDialogId is not None ) and
              (self._stateMachine.cpaLegStatus == ConnectionStatus.Connected) and
              self._stateMachine.reinvitedCpa == False):
            # re-invite cpa inorder to advertise callee sdp...
            calleeSdp =  in_event.getSdpContents()
            if self._stateMachine.reInviteDialog(self._stateMachine.cpaDialogId,
                                                 calleeSdp):
              self._stateMachine.reinvitedCpa = True
            # self._stateMachine.sendInfoMsgToDialog(self._stateMachine.cpaDialogId,
            # "application","SDP",calleeSdp.getBodyDataStringCopy())

      elif(in_event.getDialogId() == self._stateMachine.cpaDialogId and
           self._stateMachine.reinvitedCpa == True):
        self._stateMachine.reinvitedCpa = False #answer from cpa endpoint
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallInfoEvent)): #OUT
      # TODO validate that the event comes from the Cpa endpoint

      # There is a race condition where the callee sends a 183 to us and some
      # media packets to the cpa endpoint and the cpa endpoint sends us an INFO
      # message before we get the 183. In this case we ignore the INFO message
      # until we receive the 183 and then we will move to WAITING_CPA_RESULT
      # state and wait for the next INFO message.
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallConnectedEvent)):
      self._stateMachine.calleeLegStatus = ConnectionStatus.Connected
      self._stateMachine.bCalleeAnswered = True
      
      # CPASTATS
      # Store CPA statistics: connected time
      self._stateMachine.cpaStats.addCurrentTimeOM(CpaOM.OM_CPA_TIME_CONNECTED)

      if (self._stateMachine.getCallStack().hadSipError()):
        # CPA failed previously: Re-invite the callee
        out_pystrNewStateName.set("REINVITING_CALLEE")
        return True
      else:
        # Notify CPA Media UAS that the Callee is now connected using an
        # INFO msg with an application/sipfrag body
        if (not self._stateMachine.sendInfoMsgToDialog(
              self._stateMachine.cpaDialogId, "application",
              "sipfrag", "SIP/2.0 200 Ok")):
          return False
        
        out_pystrNewStateName.set("WAITING_CPA_RESULT")
        return True
    
    elif (isinstance(in_event, TimeoutEvent)):
      
      #
      # Since it is a _dialing_ state, we can assume that the timeout
      # is of pre-connect type.
      #
      
      PyLoggingFunctions.log_info(self.getLogger(),\
            "DialingCalleeState: cpa.result.preconnect.timeout.ms expired")
        
      self._stateMachine.notifyCallEvent( CallEvent.Timeout( CallEvent.Timeout.Preconnect))
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, TimeoutCancelledEvent)):
      self._preConnectTimerId = None
      return True
    
    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      return self._handleDialogToCallTerminatedEvent(in_event, out_pystrNewStateName)
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True
  
  def onStateExit(self, in_event, in_strDestStateName):
    __pychecker__ = 'unusednames=in_strDestStateName'

    PyLoggingFunctions.log_trace(self.getLogger(), "%s::onStateExit %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    if (self._preConnectTimerId is not None):
      self._stateMachine.globalContext.cancelTimer(self._preConnectTimerId)
    
    strCpdResult = self._stateMachine.getCpdResultFromStateVariables()
    
    PyLoggingFunctions.log_debug(self.getLogger(),
      "Setting CpaOM.OM_CPA_RESULT to %s." % strCpdResult)
    #Store the result in class that ends-up in Cpa-Stats.csv. Note that this
    #should be the ONLY place where CpaOM.OM_CPA_RESULT is written.
    self._stateMachine.cpaStats.addOM(CpaOM.OM_CPA_RESULT, strCpdResult)
    self._stateMachine.cpaStats.addCurrentTimeOM(CpaOM.OM_CPA_TIME_CPA_COMPLETED)
    
    specialEvent = self._stateMachine.getDetailedCpdResultFromStateVariables()    
    self._stateMachine.cpaStats.addOM(CpaOM.OM_SPECIAL_CALL_INFORMATION, specialEvent)  
    
##EndOf: DialingCalleeState

########################################################################

# Global: regex to look for in SDP for detecting T.38 FAX
sdpWithT38regex = re.compile('m=image .* t38')

class WaitForCpaResultCommonState(CpaStates):
  """
  Base class for the states that are waiting for information (CPA result)
  from the CPA.
  
  @note Abstract class.
  """
  

  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    self.__SMTimerManager = SMTimerManager.SMTimerManager(in_stateMachine)
    self.__CheckPostConnectTimerExpiryAction = False
  
  def getTimerManager(self):
    return self.__SMTimerManager

  def setPostConnectTimerExpiryAction(self, in_action):
    PyLoggingFunctions.log_debug(self.getLogger(),
                                 "%s::setPostConnectTimerExpiryAction: Setting post connect timer expiry action checking to %r",
                                 self.__class__.__name__, in_action)
    self.__CheckPostConnectTimerExpiryAction = in_action

  def getPostConnectTimerExpiryAction(self):
    return self.__CheckPostConnectTimerExpiryAction

  
  def reset(self):
    """
    Since the states are reused, the information needs to be reset in between
    calls.
    """
    
    # Make sure no there are no left-over timers
    assert self.getTimerManager().getNumTimers() == 0
    self.setPostConnectTimerExpiryAction(False)

  def _handle_TimeoutEvent_Custom(self, in_event):
    """
    This virtual function can be used by the child to provide specific 
    functionality on a timeout. The child does not necessarily have to
    call the parent.
    """
    __pychecker__ = 'unusednames=in_event'
    #Flag the call with proper error code, do not want ambiguous 'unknown'.
    
    # Here we can be sure the timeout is of Postconnect type
    self._stateMachine.notifyCallEvent( CallEvent.Timeout( CallEvent.Timeout.Postconnect ) )
  
    
  # @abstract
  def _handle_DialogToCallInfoEvent(self, in_event, out_pystrNewStateName):
    """
    Should never happen: this function is abstract: it HAS TO be overloaded
    by a child.
    """
    __pychecker__ = 'unusednames=in_event,out_pystrNewStateName'
    assert False

  def processEvent_i(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0 maxlines=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallFailedEvent)): #OUT
      self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      
      sipFailedStatusCode = in_event.getSipMessage().getStatusCode()
      self._stateMachine.notifyCallEvent( CallEvent.SipError( CallEvent.SipErrorOriginator.Callee, 
                                                              CallEvent.SipErrorCode(sipFailedStatusCode))
                                        )
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent)): #OUT

      if (in_event.getDialogId() == self._stateMachine.calleeDialogId):#SDP from callee

        if (self._stateMachine.getCallStack().hadSipError()):
          # Save the callee sdp answered
          self._stateMachine.calleeSdp = in_event.getSdpContents()
        else:

          if ((self._stateMachine.cpaDialogId is not None ) and
              (self._stateMachine.cpaLegStatus == ConnectionStatus.Connected) and
              (self._stateMachine.reinvitedCpa == False)):
            # re-invite cpa in order to advertise callee sdp...
            calleeSdp =  in_event.getSdpContents()
            if self._stateMachine.reInviteDialog(self._stateMachine.cpaDialogId,
                                              calleeSdp):
              self._stateMachine.reinvitedCpa = True
            # self._stateMachine.sendInfoMsgToDialog(self._stateMachine.cpaDialogId,"application",
            # "SDP",calleeSdp.getBodyDataStringCopy())
            
      elif(in_event.getDialogId() == self._stateMachine.cpaDialogId and
           self._stateMachine.reinvitedCpa == True):
        self._stateMachine.reinvitedCpa = False #answer from cpa endpoint
      
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallConnectedEvent)):
      self._stateMachine.calleeLegStatus = ConnectionStatus.Connected
      self._stateMachine.bCalleeAnswered = True
      
      # CPASTATS
      # Store CPA statistics: connected time
      self._stateMachine.cpaStats.addCurrentTimeOM(CpaOM.OM_CPA_TIME_CONNECTED)
      
      if (self._stateMachine.getCallStack().hadSipError()):
        # CPA failed previously: Re-invite the callee
        out_pystrNewStateName.set("REINVITING_CALLEE")
        return True
      else:
        preTimerInfo = self.getTimerManager().getTimerInfoByName("PreConnect")
        if preTimerInfo is not None:
          ## Stop preConnect timer and start postConnect timer
          
          #Store the PreConnectTime (will be lost after timer destruction)
          self._fPreConnectTime += (time.clock() - preTimerInfo.fTimeRef)
            
          self.getTimerManager().removeTimerByName("PreConnect")
        else:
          PyLoggingFunctions.log_warn(self.getLogger(),
            "Could not find the PreConnect timer in DialogToCallConnectedEvent."
            " In class %s." % self.__class__.__name__)

        # start the post connect timeout
        postConnectTimeout =  \
          self._stateMachine.config[self._stateMachine.PARAM_CPA_POSTCONNECT_TIMEOUT_MS]
        self.getTimerManager().startTimer("PostConnect", postConnectTimeout)
        if postConnectTimeout < 5000:
          self.setPostConnectTimerExpiryAction(self._stateMachine.config[self._stateMachine.PARAM_CPA_ALLOW_POST_CONNECT_TIMEOUT_DELAYING])
        else:
          self.setPostConnectTimerExpiryAction(False)

        PyLoggingFunctions.log_info(self.getLogger(),\
                                    "CPA result: call connected. PostConnectTimerExpiryAction is %r",
                                    self.getPostConnectTimerExpiryAction())

        # Notify CPA Media UAS that the Callee is now connected using an
        # INFO msg with an application/sipfrag body
        if (not self._stateMachine.sendInfoMsgToDialog(
              self._stateMachine.cpaDialogId, "application",
              "sipfrag", "SIP/2.0 200 Ok")):
          return False
        
        # Stay in the same state
        return True
    
    elif (isinstance(in_event, DialogToCallInfoEvent)):
      rv = self._handle_DialogToCallInfoEvent(in_event, out_pystrNewStateName)
      assert rv != None
      return rv
    
    elif (isinstance(in_event, DialogToCallInfoSentSucceededEvent)): #OUT
      # TODO: validate it is coming from the CPA endpoint
      # Response to the INFO with the Callee progress as application/sipfrag
      
      # TODO: if more than one event is reported (currently only 200 Ok), we'll
      #       need to make sure the response for the previous INFO was received
      #       before sending the next one
      
      # Stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallInfoSentFailedEvent)): 
      # May be ignore, there nothing we can do anyway
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, OutDialogToCallProvisionalEvent)):
      if (in_event.getDialogId() == self._stateMachine.calleeDialogId and self._stateMachine.callerDialogId is not None):
        # More provisional responses from Callee: (bug 1018)
        relay_provisional = self._stateMachine.\
                            config[self._stateMachine.PARAM_CPA_RELAY_PROVISIONAL_WHILE_WAITING_FOR_RESULT]

        if relay_provisional is True:
          PyLoggingFunctions.log_info(self.getLogger(),
                                      "WaitForCpaResultCommonState.processEvent: "
                                      "Relaying provisional response received from Callee "
                                      "while waiting for CPA result")
          statusCodeToRelay = None
          statusCode = in_event.getSipMessage().getStatusCode()
          if (statusCode == 183):
            statusCodeToRelay = self._stateMachine.config[self._stateMachine.PARAM_CPA_PROVISIONAL_RELAYED_UPON_183]
          else:
            statusCodeToRelay = statusCode
          
          self._stateMachine.sendProvisionalToInDialog(self._stateMachine.callerDialogId, statusCodeToRelay)
          
        else:
          PyLoggingFunctions.log_info(self.getLogger(),
                                      "WaitForCpaResultCommonState.processEvent: "
                                      "Discarding provisional response received from Callee "
                                      "while waiting for CPA result")
      else:
        PyLoggingFunctions.log_info(self.getLogger(),
                                    "WaitForCpaResultCommonState.processEvent: "
                                    "Discarding unexpected provisional response received "
                                    "while waiting for CPA result")
      # Stay in the same state
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent)): #OUT

      # Re-INVITE with SDP from the callee
      assert in_event.getDialogId() == self._stateMachine.calleeDialogId

      calleeSdpOffer = in_event.getSdpContents()
      strSDP = calleeSdpOffer.getBodyDataStringCopy()

      # Bug 6909: Look for t.38 FAX re-INVITEs
      if(sdpWithT38regex.search(strSDP)):

        PyLoggingFunctions.log_debug(self.getLogger(),
            "Re-INVITE from callee with T.38 media line (FAX)")

        # Set the call CPA result as FAX
        self._stateMachine.notifyCallEvent( 
          CallEvent.Factory_CallCpaResult(CpaResult.CpaResult.Fax) )

        # Transition to T.38 FAX handling state
        out_pystrNewStateName.set("T38FAX")
        return True

      else:
        # Bug 3391: re-INVITEd by callee - but not for T.38 FAX - warn and 
        #   reject the re-INVITE with "403 Forbidden"
        PyLoggingFunctions.log_warn(self.getLogger(),
            "Re-INVITE received from media gateway/SIP trunk while performing"
            " CPA analysis - rejecting using \"403 Forbidden\" response")
        if (not self._stateMachine.rejectDialog(
              self._stateMachine.calleeDialogId, 
              403)):
          return False
        # Stay in the same state
        return True

    elif (isinstance(in_event, TimeoutEvent)):

      PyLoggingFunctions.log_debug(self.getLogger(), "TimeoutEvent in: " + self.__class__.__name__)

      # check whether this tiemout event is intended to stop cpa detection
      stop_cpa = True
      timerInfo = self.getTimerManager().getTimerInfoByID(in_event.timerId)
      if (timerInfo.strTimerTag == "PostConnect" and self.getPostConnectTimerExpiryAction() is True):
        PyLoggingFunctions.log_debug(self.getLogger(), "Checking post connect timeout processing action with engine.")
        stop_cpa = False
      

      #PyLoggingFunctions.log_debug(self.getLogger(), "stop_cpa %r", stop_cpa)

      if (self._stateMachine.calleeLegStatus == ConnectionStatus.Connected):
        if stop_cpa is False:
          # post connect timeout expires but we need to query for CPA
          # endpoint to see if we need to delay the cpa endpoint
          # disconnection

          # send info message to cpa endpoint to query for action to
          # perform...
          if (self._stateMachine.cpaDialogId is None or
              not self._stateMachine.sendInfoMsgToDialog(self._stateMachine.cpaDialogId,
                                                         "application", "sipfrag",
                                                         "CALL_CONTROL_EVENT post_connect_timer_expiration")):
            stop_cpa = True

        if stop_cpa is True:
          self._handle_TimeoutEvent_Custom(in_event)
          # Hanging up the cpa endpoint
          PyLoggingFunctions.log_info(self.getLogger(), "hanging up cpa endpoint")
          if (not self._stateMachine.hangupDialog(self._stateMachine.cpaDialogId)):
            PyLoggingFunctions.log_error(self.getLogger(), "Failed to hangup cpa endpoint")
            return False
          
          self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnecting
          out_pystrNewStateName.set("REINVITING_CALLEE")
        else:
          # remove all timers...
          #self.getTimerManager().removeAllTimers()
          # start a one second timer to wait for answer to the call control event we just sent...
          PyLoggingFunctions.log_info(self.getLogger(), "Post Connect timeout expired, but checking with engine for action to perform.")
          self.getTimerManager().startTimer("WaitForCCEventAnswer", 1000)
          
        return True
      else:
        self._handle_TimeoutEvent_Custom(in_event)
        PyLoggingFunctions.log_debug(self.getLogger(), "callee leg not connected. Going to disconnect the call.")
        out_pystrNewStateName.set("DISCONNECTING")
        return True

    elif (isinstance(in_event, TimeoutCancelledEvent)):
      self.getTimerManager().removeTimerByID(in_event.timerId)
      return True
    
    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      return self._handleDialogToCallTerminatedEvent(in_event, out_pystrNewStateName)
      
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True
 
  def onStateExit(self, in_event, in_strDestStateName):
    __pychecker__ = 'unusednames=in_strDestStateName'
    PyLoggingFunctions.log_debug(self.getLogger(), "%s::onStateExit %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    #Add custom call information
    specialEvent = self._stateMachine.getDetailedCpdResultFromStateVariables()
    if specialEvent is None:
      specialEvent = DetailedCpdResult.DetailedCpdResult.NoResult
    PyLoggingFunctions.log_debug(self.getLogger(),
      "Setting CpaOM.OM_SPECIAL_CALL_INFORMATION to %s." % specialEvent)
    self._stateMachine.cpaStats.addOM(CpaOM.OM_SPECIAL_CALL_INFORMATION, specialEvent)    
    
    PyLoggingFunctions.log_debug(self.getLogger(), "Cancelling unused timer(s)")
    self.getTimerManager().removeAllTimers()
##EndOf: WaitForCpaResultCommonState
  
########################################################################

class WaitingForCpaPreOrPostResultState(WaitForCpaResultCommonState): 
  def __init__(self, in_stateMachine):
    super(WaitingForCpaPreOrPostResultState,self).__init__(in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "WaitingForCpaPreOrPostResultState.__init__")
    
    self._fPreConnectTime = 0.0    
    self._fPostConnectTime = 0.0   
  
  def reset(self):
    self._fPreConnectTime = 0.0    
    self._fPostConnectTime = 0.0
    super(WaitingForCpaPreOrPostResultState,self).reset()
  
  def getName(self):
    return "WAITING_CPA_RESULT"  
  
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    self.reset()
    
    if (self._stateMachine.calleeLegStatus == ConnectionStatus.Connected):
      uiTimeoutMs = \
        self._stateMachine.config[self._stateMachine.PARAM_CPA_POSTCONNECT_TIMEOUT_MS]
      self.getTimerManager().startTimer("PostConnect", uiTimeoutMs)

      if uiTimeoutMs < 5000:
        self.setPostConnectTimerExpiryAction(self._stateMachine.config[self._stateMachine.PARAM_CPA_ALLOW_POST_CONNECT_TIMEOUT_DELAYING])
      else:
        self.setPostConnectTimerExpiryAction(False)

      PyLoggingFunctions.log_info(self.getLogger(),\
                                  "CPA result: call connected.PostConnectTimerExpiryAction is %r",
                                  self.getPostConnectTimerExpiryAction())
    else:
      uiTimeoutMs = \
        self._stateMachine.config[self._stateMachine.PARAM_CPA_PRECONNECT_TIMEOUT_MS]
      self.getTimerManager().startTimer("PreConnect", uiTimeoutMs)
      PyLoggingFunctions.log_info(self.getLogger(),\
                                  "CPA result: call in early media")
    
    out_pystrNewStateName.set(self.getName())
  
  def _handle_TimeoutEvent_Custom(self, in_event):
    WaitForCpaResultCommonState._handle_TimeoutEvent_Custom(self, in_event)
    
    timerInfo = self.getTimerManager().getTimerInfoByID(in_event.timerId)
    
    if timerInfo.strTimerTag == "PreConnect":
      #Can this ever happen (this state cannot be in pre-connect)!?
      PyLoggingFunctions.log_info(self.getLogger(),\
                    "WaitingForCpaPreOrPostResultState: cpa.result.preconnect.timeout.ms expired")
      
      self._stateMachine.notifyCallEvent( CallEvent.Timeout( CallEvent.Timeout.Preconnect))
    elif timerInfo.strTimerTag == "PostConnect":
      PyLoggingFunctions.log_info(self.getLogger(),\
                    "WaitingForCpaPreOrPostResultState: cpa.result.postconnect.timeout.ms expired")
      
      self._stateMachine.notifyCallEvent( CallEvent.Timeout( CallEvent.Timeout.Postconnect))
    elif (timerInfo.strTimerTag == "WaitForCCEventAnswer" or timerInfo.strTimerTag == "WaitForAMresult"):
      self._stateMachine.notifyCallEvent( CallEvent.Timeout( CallEvent.Timeout.Postconnect))
      PyLoggingFunctions.log_info(self.getLogger(),\
                    "WaitingForCpaPreOrPostResultState: %s expired", timerInfo.strTimerTag)
    else:
      PyLoggingFunctions.log_warn(self.getLogger(), "Unknown timeout expired.")


  def _handle_DialogToCallInfoEvent(self, in_event, out_pystrNewStateName):
    if (in_event.getDialogId() == self._stateMachine.cpaDialogId):
      # INFO message received from the CPA Media endpoint (out leg): check if 
      # it's coming with intermediate results (application/cpa-update MIME type)
      sipMsg = in_event.getSipMessage()
      
      PyLoggingFunctions.log_debug(\
          self.getLogger(),\
          "CPA state 'WaitingForCpaPreOrPostResultState' got SIP INFO message with body:\n" +\
          sipMsg.getBodyText() )

      if sipMsg.hasContents() and sipMsg.getContentType() == "application" and \
         sipMsg.getContentSubType() == "cpa-update":
        
        if(sipMsg.getBodyText() == "CPA_CC_DELAY_HANGUP" or sipMsg.getBodyText() == "CPA_CC_PROCESS_HANGUP"):
          self.getTimerManager().removeTimerByName("WaitForCCEventAnswer")
          ret_value = True
          
          if sipMsg.getBodyText() == "CPA_CC_DELAY_HANGUP":
            # give three more seconds to cpa endpoint to try converging towards AM
            self.getTimerManager().startTimer("WaitForAMresult",3000)
            PyLoggingFunctions.log_info(self.getLogger(),"Waiting a little more to get post connect result.")
          else:
            # Hanging up the cpa endpoint
            if (not self._stateMachine.hangupDialog(self._stateMachine.cpaDialogId)):
              PyLoggingFunctions.log_eror(self.getLogger(),"Failed to hangup cpa dialog")
              ret_value = False
            else:
              self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnecting
              
            out_pystrNewStateName.set("REINVITING_CALLEE")
          return ret_value

      
      #
      # Wait for an acceptable CPA result. Reject the messages 
      # where the threshold has not been attained.
      #
      # Note: Only get a CPA Result from the CPA if its result 
      # is > threshold set in options.
      strPrePostOutcome = self._stateMachine.getFinalCPAOutcomeFromInfoMsg(
        sipMsg,
        in_bCalleeIsConnected = \
            self._stateMachine.calleeLegStatus == ConnectionStatus.Connected)
      
      # Have everything we want, the Cpa can ask to terminate the call and 
      # let the session controller initiate the call in between the caller 
      # and callee.
      if strPrePostOutcome is not None:
        if not CpaResult.isPreOrPostConnect(strPrePostOutcome):
          PyLoggingFunctions.log_warn(
                         self.getLogger(),
                         "Received invalid result (not PreOrPostConnect)"
                         " from Cpa: %s." % strPrePostOutcome)
        else:
          PyLoggingFunctions.log_info(
                         self.getLogger(), 
                         "Received CPA result of %s; reintegrating call flow." \
                         % strPrePostOutcome)
          
          assert CpaResult.isPreOrPostConnect(strPrePostOutcome)
          
          #
          # Log the amount of time it took for detection
          #
          if (self._stateMachine.calleeLegStatus == ConnectionStatus.Connected):
            postTimerInfo = self.getTimerManager().getTimerInfoByName("PostConnect")
            self._fPostConnectTime += (time.clock() - postTimerInfo.fTimeRef)
          else:
            preTimerInfo = self.getTimerManager().getTimerInfoByName("PreConnect")
            self._fPreConnectTime += (time.clock() - preTimerInfo.fTimeRef)
          
          PyLoggingFunctions.log_info(self.getLogger(), 
                                     "CPA result: final result %s at %4.2f "
                                     "second(s) after connection with %4.2f "
                                     "second(s) of pre-connect analysis",
                   strPrePostOutcome, self._fPostConnectTime, self._fPreConnectTime)
          
          #Update CpaOutcome
          self._stateMachine.notifyCallEvent( CallEvent.Factory_CallCpaResult(strPrePostOutcome) )
          
          #Call virtual function, the transition depends on whether or not the Cpa
          #object is an 'Eamd' object or not.
          return self._stateMachine.transitionFromPreOrPostDetectionDone(out_pystrNewStateName)

    #Received an info message not from the NCA Engine
    sipMsg = in_event.getSipMessage()
    
    #Exception for a message from the SIP Media Server
    if not "MEDIA-RESULT: END" in sipMsg.getBodyText():
      PyLoggingFunctions.log_warn(self.getLogger(), "Received an unexpected INFO message "
                                                    "in current state (%s). Sip message " 
                                                    "received: %s." % (self.getName(), sipMsg.getBodyText()))
    return True
  
  def onStateExit(self, in_event, in_strDestStateName):
    __pychecker__ = 'unusednames=in_strDestStateName'
    
    PyLoggingFunctions.log_debug(self.getLogger(), "%s::onStateExit %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    callEvents = self._stateMachine.getCallStack()
    cpaResult = callEvents.getCpaResult()
    
    if (cpaResult is None):
   
      postTimerInfo = self.getTimerManager().getTimerInfoByName("PostConnect")
      preTimerInfo = self.getTimerManager().getTimerInfoByName("PreConnect")
      
      if postTimerInfo is not None:
        self._fPostConnectTime = (time.clock() - postTimerInfo.fTimeRef)
      elif preTimerInfo is not None:
        self._fPreConnectTime = (time.clock() - preTimerInfo.fTimeRef)
      else:
        PyLoggingFunctions.log_error(\
              self.getLogger(), 
              "Incoherent timers, should have at least 1 "
              "{pre,post} connect timer.")
      
      PyLoggingFunctions.log_info(self.getLogger(),
               "CPA result: stopped without any final result "
               "%4.2f second(s) after connection with %4.2f second(s) of "
               "pre-connect analysis",
               self._fPostConnectTime, self._fPreConnectTime)
    
    strCpdResult = self._stateMachine.getCpdResultFromStateVariables()
                                                          
    PyLoggingFunctions.log_debug(self.getLogger(),
      "Setting CpaOM.OM_CPA_RESULT to %s." % strCpdResult)
    #Store the result in class that ends-up in Cpa-Stats.csv. Note that this
    #should be the ONLY place where CpaOM.OM_CPA_RESULT is written.
    self._stateMachine.cpaStats.addOM(CpaOM.OM_CPA_RESULT, strCpdResult)
    self._stateMachine.cpaStats.addCurrentTimeOM(CpaOM.OM_CPA_TIME_CPA_COMPLETED)
    specialEvent = self._stateMachine.getDetailedCpdResultFromStateVariables()
    self._stateMachine.cpaStats.addOM(CpaOM.OM_SPECIAL_CALL_INFORMATION, specialEvent)  
    
    super(WaitingForCpaPreOrPostResultState,self).onStateExit(in_event, 
          in_strDestStateName)

## EndOf: WaitingForCpaPreOrPostResultState

class T38FaxState(CpaStates):
  """
  A re-invite with T38 FAX was received from Callee (bug 6909)
  """
  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "T38FaxState.__init__")
  
  def getName(self):
    return "T38FAX"
    
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())

    # Reject re-INVITE for T.38 with 415 Unsupported media type
    if (not self._stateMachine.rejectDialog(
          self._stateMachine.calleeDialogId, 
          415)):
      return False

    # Hangup callee
    if (not self._stateMachine.hangupDialog(
          self._stateMachine.calleeDialogId)):
      return False
    self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting

    # Answer the caller with a dummy SDP and "CPD-Result: Fax" header
    # NOTE: mu-law only
    errStr = PyString()
    dummySdpAnswer = SDP()
    strSDP = "v=0\r\n" \
        "o=- 1182439127 1182439127 IN IP4 127.0.0.1\r\n" \
        "s=NCA Dummy Media Session for T.38 FAX\r\n" \
        "c=IN IP4 127.0.0.1\r\n" \
        "t=0 0\r\n" \
        "m=audio 9000 RTP/AVP 0\r\n" \
        "a=sendonly\r\n" \
        "a=rtpmap:0 PCMU/8000\r\n"

    if not dummySdpAnswer.parse(strSDP, errStr):
      PyLoggingFunctions.log_error(self.getLogger(), \
          "failed to parse sdp :\n%s", errStr)
      return False
      
    if (not self._stateMachine.answerInDialog(
          self._stateMachine.callerDialogId,
          dummySdpAnswer,
          self._stateMachine.callerSdp is None, # Must set to True if SDP is offer
          self._stateMachine.getAnswerMessageWithCpaResult())):
      return False

    self._stateMachine.callerLegStatus = ConnectionStatus.Answering

    # Stay in same state - waiting for the caller to hang up
    return True

  
  def processEvent_i(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
   
    # Expecting:
    # - DialogToCallTerminatedEvent from callee
    # - DialogToCallConnectedEvent from caller
    # - DialogToCallTerminatedEvent from caller (hangs up on FAX disposition)

    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())

    if (isinstance(in_event, DialogToCallTerminatedEvent)):
      
      if (in_event.getDialogId() == self._stateMachine.calleeDialogId):
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnected
        return True

      elif (in_event.getDialogId() == self._stateMachine.callerDialogId):
        # We are done with this call
        self._stateMachine.callerLegStatus = ConnectionStatus.Disconnected      
        out_pystrNewStateName.set("DISCONNECTING")
        return True

    elif (isinstance(in_event, DialogToCallConnectedEvent)):
      pass

    return True

##EndOf: T38FaxState 

########################################################################

class ReinvitingCalleeState(CpaStates):
  """
  Connects RTP streams for Caller and Callee together once CPA is done
  """
  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "ReinvitingCalleeState.__init__")
  
  def getName(self):
    return "REINVITING_CALLEE"
    
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)

    self._calleeGlareDetected = False
    
    # self._stateMachine.callerSdp can be None (handled differently 
    #  by reInvite)
    if (not self._stateMachine.reInviteDialog(
          self._stateMachine.calleeDialogId, self._stateMachine.callerSdp)):
      return
    
    out_pystrNewStateName.set(self.getName())
  
  def processEvent_i(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallFailedEvent) and (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # Callee refused the re-INVITE with the new SDP
      sipFailedStatusCode = in_event.getSipMessage().getStatusCode()
      #check is this is due to a glare condition
      if (sipFailedStatusCode == 491):
        PyLoggingFunctions.log_info(self.getLogger(), "Glare detected while reinviting callee. Handling received re-INVITE from callee.\n")
        self._calleeGlareDetected = True

      else:
        # When re-INVITE fails we must not set the connection status to
        # disconnecting
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      
        self._stateMachine.notifyCallEvent( CallEvent.SipError( CallEvent.SipErrorOriginator.Callee, 
                                          CallEvent.SipErrorCode(sipFailedStatusCode))
                                         )
        out_pystrNewStateName.set("DISCONNECTING")
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent) and (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # Save the callee sdp answered
      self._stateMachine.calleeSdp = \
          in_event.getSdpContents()
      
      # Answer the caller with the callee sdp
      if (not self._stateMachine.answerInDialog(
            self._stateMachine.callerDialogId,
            self._stateMachine.calleeSdp,
            False, # SDP is answer
            self._stateMachine.getAnswerMessageWithCpaResult())):
        return False
      self._stateMachine.callerLegStatus = ConnectionStatus.Answering
      
      out_pystrNewStateName.set("ANSWERING_CALLER")
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent) and (in_event.getDialogId() == self._stateMachine.calleeDialogId)): #OUT
      # Save the callee sdp offer (caller sent INVITE w/o sdp and/or glare during reinvite)
      calleeFinalSdpOffered = in_event.getSdpContents()
      callerSdpIsOffer = True

      if (self._calleeGlareDetected and self._stateMachine.callerSdp):
        # answer sdp offer from
        if (not self._stateMachine.provideSdpAnswerToDialog(self._stateMachine.calleeDialogId,self._stateMachine.callerSdp)):
          return False

        # Tag the answer to the caller as an SDP answer
        callerSdpIsOffer = False
        self._calleeGlareDetected = False
    
      
      # Use the callee sdp offer as the sdp offer for the caller
      if (not self._stateMachine.answerInDialog(
            self._stateMachine.callerDialogId,
            calleeFinalSdpOffered,
            callerSdpIsOffer, # SDP is offer
            self._stateMachine.getAnswerMessageWithCpaResult())):
        return False
      
      self._stateMachine.callerLegStatus = ConnectionStatus.Answering
      
      out_pystrNewStateName.set("ANSWERING_CALLER")
      return True
    
    elif (isinstance(in_event, DialogToCallInfoSentSucceededEvent)): #OUT
      # TODO: validate it is coming from the CPA endpoint
      # Response to the INFO with the Callee progress as application/sipfrag      
      #   We don't care at this point: stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallInfoEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      return self._handleDialogToCallTerminatedEvent(in_event, out_pystrNewStateName)
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True

########################################################################

class AnsweringCallerState(CpaStates):
  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "AnsweringCallerState.__init__")
  
  def getName(self):
    return "ANSWERING_CALLER"
  
  def processEvent_i(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0 maxlines=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallConnectedEvent)):

      if (in_event.getDialogId() == self._stateMachine.callerDialogId):
        self._stateMachine.callerLegStatus = ConnectionStatus.Connected
      
        # CPASTATS
        # Store CPA statistics: Agent answered
        if( not self._stateMachine.cpaStats.isOM(CpaOM.OM_CPA_TIME_QUEUED) ):
          # If call was not queued then consider the agent connected
          self._stateMachine.cpaStats.addCurrentTimeOM(CpaOM.OM_CPA_TIME_AGENT_CONNECTED)

        out_pystrNewStateName.set("CALLER_CALLEE_CONNECTED")
        return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent) and (in_event.getDialogId() == self._stateMachine.callerDialogId)): #IN
      # Send the caller sdp answer (in ACK) as the callee sdp answer (in ACK)
      # (caller sent INVITE w/o sdp)
      if (not self._stateMachine.provideSdpAnswerToDialog(
            self._stateMachine.calleeDialogId,
            in_event.getSdpContents())):
        return False

      # CPASTATS
      if(self._stateMachine.isSdpFromStreamManager(in_event.getSdpContents())):
        # Callee is being queued before talking to an agent, just keep track of
        #   it in the stats
        self._stateMachine.cpaStats.addCurrentTimeOM(CpaOM.OM_CPA_TIME_QUEUED)

      # Do not change state: DialogToCallConnectedEvent coming just after
      return True
    
    elif (isinstance(in_event, DialogToCallAckNotReceivedEvent) and (in_event.getDialogId() == self._stateMachine.callerDialogId)): #IN
      # Can happen here after answering 200 Ok to an INVITE without SDP, 
      #   see bug 989
      # TODO use mutable logger for this warning
      PyLoggingFunctions.log_warn(self.getLogger(), "The SIP UAC did not answer our " \
        "SIP 200 response with a ACK request as specified by RFC 3261. " \
        "Terminating the call. (State: %s)",self.getName())
      
      out_pystrNewStateName.set("DISCONNECTING")
      return True
      
    elif (isinstance(in_event, DialogToCallInfoEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      return self._handleDialogToCallTerminatedEvent(in_event, out_pystrNewStateName)
    
    elif (isinstance(in_event, StopEvent)):
      out_pystrNewStateName.set("DISCONNECTING")
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True


class CallerCalleeConnectedState(B2buaStates.CallerCalleeConnectedState):
  def __init__(self, in_stateMachine):
    B2buaStates.CallerCalleeConnectedState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "CallerCalleeConnectedState.__init__")
  
  def processEvent_i(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallTerminatedEvent)):
      return self._handleDialogToCallTerminatedEvent(in_event, out_pystrNewStateName)
    
    else:
      # processEvent from based class (B2BUA)
      bRV = B2buaStates.CallerCalleeConnectedState.processEvent(
        self, in_event, out_pystrNewStateName)
      
      return bRV
    
    return True


class ReinvitedState(B2buaStates.ReinvitedState):
  """
  Re-invite received from caller once cpa is over and caller and callee streams
  are connected.
  """
  def __init__(self, in_stateMachine):
    B2buaStates.ReinvitedState.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "ReinvitedState.__init__")
  
  def processEvent_i(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, DialogToCallTerminatedEvent)):
      return self._handleDialogToCallTerminatedEvent(in_event, out_pystrNewStateName)
    
    else:
      # processEvent from based class (B2BUA)
      return B2buaStates.ReinvitedState.processEvent(
        self, in_event, out_pystrNewStateName)
    
    return True


class DisconnectingState(CpaStates):
  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DisconnectingState.__init__")
  
  def getName(self):
    return "DISCONNECTING"
  
  def entryAction(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::entryAction %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    
      
    if (self._stateMachine.dialogIdInitiatingDisconnect is not None):
      if (self._stateMachine.dialogIdInitiatingDisconnect == \
          self._stateMachine.callerDialogId):
        self._stateMachine.callerLegStatus = ConnectionStatus.Disconnected
      elif (self._stateMachine.dialogIdInitiatingDisconnect == \
            self._stateMachine.calleeDialogId):
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnected
      else:
        self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnected
    
    if (self._stateMachine.callerLegStatus == ConnectionStatus.Disconnected and \
        (self._stateMachine.calleeLegStatus == ConnectionStatus.Unknown or \
         self._stateMachine.calleeLegStatus == ConnectionStatus.Disconnected) and \
        (self._stateMachine.cpaLegStatus == ConnectionStatus.Unknown or \
         self._stateMachine.cpaLegStatus == ConnectionStatus.Disconnected)):
      out_pystrNewStateName.set("DISCONNECTED")
      return
    
    # Reject the caller if it's connecting
    if (self._stateMachine.callerLegStatus == ConnectionStatus.Connecting):

      PyLoggingFunctions.log_info(self.getLogger(), 'Rejecting caller cpa')
      
      rejectResponseCode, rejectSipMessage = \
        self._stateMachine.getRejectMessageAndSipErrorCodeFromStateVariables()
      
      PyLoggingFunctions.log_debug(self.getLogger(), 
        'Rejecting caller cpa with: rejectResponseCode: %s, rejectSipMessage: %s' %
        (rejectResponseCode, rejectSipMessage))
    
      PyLoggingFunctions.log_debug(self.getLogger(),
        "Call state: " % self._stateMachine.getCallStack() )
      
      if (not self._stateMachine.rejectDialog(
            self._stateMachine.callerDialogId,
            rejectResponseCode,
            rejectSipMessage)):
        return
      
      self._stateMachine.callerLegStatus = ConnectionStatus.Disconnecting
    
    # Hanging up the caller if it's connected
    elif (self._stateMachine.callerLegStatus == ConnectionStatus.Answering or
          self._stateMachine.callerLegStatus == ConnectionStatus.Connected):
      if (not self._stateMachine.hangupDialog(
            self._stateMachine.callerDialogId)):
        return
      self._stateMachine.callerLegStatus = ConnectionStatus.Disconnecting

    # Hanging up the cpa if it's connecting or connected.If its
    # connected, and we already issued a re-invite to the cpa
    # endpoint, wait first for the upcoming 200 Ok response for the
    # sent response before disconnecting the cpa..

    if ((self._stateMachine.cpaLegStatus == ConnectionStatus.Connecting) or
        (self._stateMachine.cpaLegStatus == ConnectionStatus.Connected and
         self._stateMachine.reinvitedCpa == False)):
      cpaSM = self._stateMachine.getOutDialog(self._stateMachine.cpaDialogId)
      if (cpaSM is not None):
        PyLoggingFunctions.log_debug(self.getLogger(), 'Disconnecting cpa leg with dialog id: %s' % self._stateMachine.cpaDialogId) 
        if (not self._stateMachine.hangupDialog(
              self._stateMachine.cpaDialogId)):
          return
        self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnecting
      else:
        # Between a new outgoing call request and the call to addOutDialog
        # function it is possible to have events on another leg, and then
        # here it is possible that getOutDialog fails because addOutDialog
        # function has not been called yet
        self._stateMachine.cpaLegStatus = ConnectionStatus.MustDisconnect

    # Hanging up the callee if it's connecting or connected
    if (self._stateMachine.calleeLegStatus == ConnectionStatus.Connecting or
        self._stateMachine.calleeLegStatus == ConnectionStatus.Connected):
      calleeSM = self._stateMachine.getOutDialog(self._stateMachine.calleeDialogId)
      if (calleeSM is not None):
        PyLoggingFunctions.log_debug(self.getLogger(), 'Disconnecting calee leg with dialog id: %s' % self._stateMachine.calleeDialogId) 
        if (not self._stateMachine.hangupDialog(
              self._stateMachine.calleeDialogId)):
          return
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      else:
        # Between a new outgoing call request and the call to addOutDialog
        # function it is possible to have events on another leg, and then
        # here it is possible that getOutDialog fails because addOutDialog
        # function has not been called yet
        self._stateMachine.calleeLegStatus = ConnectionStatus.MustDisconnect
  
  
    dialogNleg = self._stateMachine.getDescriptionOfDialogAndLegStatus()
    PyLoggingFunctions.log_debug(self.getLogger(), "DialogAndLegStatus after entryAction: \n " + dialogNleg)
    
  def processEvent_i(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'maxreturns=0'
    
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    if (isinstance(in_event, OutDialogToCallCallingEvent)):
      if (in_event.getDialogId() == self._stateMachine.calleeDialogId and \
          self._stateMachine.calleeLegStatus == ConnectionStatus.MustDisconnect):
        # Hanging up the callee
        if (not self._stateMachine.hangupDialog(
              self._stateMachine.calleeDialogId)):
          return False
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      
      elif (in_event.getDialogId() == self._stateMachine.cpaDialogId and \
            self._stateMachine.cpaLegStatus == ConnectionStatus.MustDisconnect):
        # Hanging up the cpa
        if (not self._stateMachine.hangupDialog(
              self._stateMachine.cpaDialogId)):
          return False
        self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnecting
      
      # Stay in the same state; DialogToCallTerminatedEvent will follow
      return True
    
    elif (isinstance(in_event, DialogToCallFailedEvent)): # OUT
      
      if (in_event.getDialogId() == self._stateMachine.calleeDialogId):
        if (self._stateMachine.calleeLegStatus == ConnectionStatus.MustDisconnect):
          self._stateMachine.hangupDialog(self._stateMachine.calleeDialogId)
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnecting
      elif (in_event.getDialogId() == self._stateMachine.cpaDialogId):
        self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnecting
      
      # Stay in the same state; DialogToCallTerminatedEvent will follow
      return True
    
    elif (isinstance(in_event, DialogToCallSdpAnsweredEvent)):

      if((in_event.getDialogId() == self._stateMachine.cpaDialogId ) and
         (self._stateMachine.cpaLegStatus == ConnectionStatus.Connected) and
         (self._stateMachine.reinvitedCpa == True)):
        self._stateMachine.reinvitedCpa = False
        cpaSM = self._stateMachine.getOutDialog(self._stateMachine.cpaDialogId)
        if (cpaSM is not None):
          self._stateMachine.hangupDialog(self._stateMachine.cpaDialogId)
          self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnecting
        else:
          # Between a new outgoing call request and the call to addOutDialog
          # function it is possible to have events on another leg, and then
          # here it is possible that getOutDialog fails because addOutDialog
          # function has not been called yet
          self._stateMachine.cpaLegStatus = ConnectionStatus.MustDisconnect
      
      # Already been cancelled in entryAction function. This event
      # was probably already pending in the queue. Then stay in the same state
      # since DialogToCallTerminatedEvent will follow.
      return True
    
    elif (isinstance(in_event, DialogToCallSdpOfferedEvent)):
      # Already been cancelled in entryAction function. This event
      # was probably already pending in the queue. Then stay in the same state
      # since DialogToCallTerminatedEvent will follow.
      return True
    
    elif (isinstance(in_event, OutDialogToCallProvisionalEvent) or \
          isinstance(in_event, DialogToCallConnectedEvent) or \
          isinstance(in_event, OutDialogToCallEarlyMediaEvent)):
      # Out call was already cancelled in entryAction function. This event
      # was probably already pending in the queue. Then stay in the same state
      # since DialogToCallTerminatedEvent will follow.
      return True
    
    elif (isinstance(in_event, DialogToCallInfoEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, DialogToCallTerminatedEvent)):
      if (in_event.getDialogId() == self._stateMachine.callerDialogId):
        self._stateMachine.callerLegStatus = ConnectionStatus.Disconnected
      elif (in_event.getDialogId() == self._stateMachine.calleeDialogId):
        self._stateMachine.calleeLegStatus = ConnectionStatus.Disconnected
      else:
        self._stateMachine.cpaLegStatus = ConnectionStatus.Disconnected

      if (self._stateMachine.callerLegStatus == ConnectionStatus.Disconnected and \
          (self._stateMachine.calleeLegStatus == ConnectionStatus.Unknown or \
           self._stateMachine.calleeLegStatus == ConnectionStatus.Disconnected) and \
          (self._stateMachine.cpaLegStatus == ConnectionStatus.Unknown or \
           self._stateMachine.cpaLegStatus == ConnectionStatus.Disconnected)):
        out_pystrNewStateName.set("DISCONNECTED")
      
      return True
    
    elif (isinstance(in_event, StopEvent)):
      # Stay in the same state
      return True

    elif (isinstance(in_event, KillEvent)):
      # Go directly in final state
      out_pystrNewStateName.set("DISCONNECTED")
      return True

    elif (isinstance(in_event, DialogToCallInfoSentSucceededEvent)):
      # Stay in the same state (can happen if in a race conditions)
      return True

    else:
      PyLoggingFunctions.log_warn(self.getLogger(),
               "%s::processEvent received an invalid event %s",
               self.__class__.__name__, in_event.__class__.__name__)
    
    return True


class DisconnectedState(CpaStates):
  def __init__(self, in_stateMachine):
    CpaStates.__init__(self, in_stateMachine)
    
    __pychecker__ = 'unusednames=t'
    t = TraceScope(self.getLogger(), "DisconnectedState.__init__")
  
  def getName(self):
    return "DISCONNECTED"
  
  # CPASTATS
  def entryAction(self, in_event, out_pystrNewStateName):
    __pychecker__ = 'unusednames=in_event,t'
    t = TraceScope(self.getLogger(),
          "DisconnectedState.entryAction")
    
    out_pystrNewStateName.set(self.getName())
    
    #@tag Review
    #Perhaps this should be the only place where the addOM is called
    #for [Detailed]CpdResult: the call should (always?) contain all
    #the information needed in the callstack.
    if not self._stateMachine.cpaStats.isOM(CpaOM.OM_CPA_RESULT):
      strCpdResult = self._stateMachine.getCpdResultFromStateVariables()
      self._stateMachine.cpaStats.addOM(CpaOM.OM_CPA_RESULT, strCpdResult)
    if not self._stateMachine.cpaStats.isOM(CpaOM.OM_SPECIAL_CALL_INFORMATION):
      specialEvent = self._stateMachine.getDetailedCpdResultFromStateVariables()
      self._stateMachine.cpaStats.addOM(CpaOM.OM_SPECIAL_CALL_INFORMATION, specialEvent)
    
    self._stateMachine.dumpCpaStats();

  def processEvent_i(self, in_event, out_pystrNewStateName):
    PyLoggingFunctions.log_trace(self.getLogger(), "%s::processEvent %s",
              self.__class__.__name__, in_event.__class__.__name__)
    
    out_pystrNewStateName.set(self.getName())
    
    # No event should be received in this state
    PyLoggingFunctions.log_warn(self.getLogger(),
             "%s::processEvent received an invalid event %s",
             self.__class__.__name__, in_event.__class__.__name__)
    
    return True
## !!MIRROR FILE, DO NOT MODIFY!!
